#include <algorithm>
#include <GL/glew.h>

#include <maya/MGlobal.h>
#include <maya/M3dView.h>
#include <maya/MTypeId.h>
#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>
#include <maya/MArrayDataHandle.h>
#include <maya/MPlug.h>
#include <maya/MFnNumericAttribute.h>
#include <maya/MFnCompoundAttribute.h>
#include <maya/MFnGenericAttribute.h>
#include <maya/MFnMessageAttribute.h>
#include <maya/MFnEnumAttribute.h>
#include <maya/MFnMesh.h>
#include <maya/MPlugArray.h>
#include <maya/MDagPath.h>
#include <maya/MFnDagNode.h>

#include "gpuCacher.h"
#include "../pluginMain.h"

MTypeId gpuCacher::id(ID_GPUCACHER);
MObject gpuCacher::aInBaseAttr;
MObject gpuCacher::aInData;
	MObject gpuCacher::aInAttr;
	MObject gpuCacher::aInMesh;
	MObject gpuCacher::aColor;
MObject gpuCacher::aInTime;
MObject gpuCacher::aExclusiveCamera;
MObject gpuCacher::aSceneStateTracking;
MObject gpuCacher::aReset;
MObject gpuCacher::aCull;
MObject gpuCacher::aSpecClrR;
MObject gpuCacher::aSpecClrG;
MObject gpuCacher::aSpecClrB;
MObject gpuCacher::aIsFrameCached;

gpuCacher::gpuCacher(){}

gpuCacher::~gpuCacher()
{
	if(_glew) vbo_delete();
	reset();
}

void* gpuCacher::creator()
{
	return new gpuCacher();
}

void gpuCacher::postConstructor()
{
	_nGeoCnt = _nIdsCnt = 0;
	_cull = true;

	if(MGlobal::mayaState() == MGlobal::kInteractive)
	{
		_glew = true;
		if(glewInit() != GLEW_OK)
		{
			cout<<"Cannot initialize GLEW."<<endl;
			_glew = false;
		}
		vbo_create();
	}
	else _glew = false;
}

MStatus gpuCacher::initialize()
{
	MFnNumericAttribute nAttr;
	MFnCompoundAttribute cAttr;
	MFnGenericAttribute gAttr;
	MFnMessageAttribute mAttr;
	MFnEnumAttribute eAttr;

	aInBaseAttr = nAttr.create("inBaseAttr", "inBaseAttr", MFnNumericData::kFloat);
	nAttr.setKeyable(true);
	nAttr.setArray(true);

	aInAttr = nAttr.create("inAttr", "inAttr", MFnNumericData::kFloat);
	nAttr.setArray(true);
	nAttr.setStorable(false);
	nAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aInMesh = gAttr.create("inMesh", "inMesh"); gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setArray(true);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aColor = nAttr.create("color", "color", MFnNumericData::k4Double);
	nAttr.setDefault(0.4, 0.4, 0.4, 1.0);

	aInData = cAttr.create("inData", "inData");
	cAttr.addChild(aInAttr);
	cAttr.addChild(aInMesh);
	cAttr.addChild(aColor);
	cAttr.setArray(true);

	aInTime = nAttr.create("inTime", "inTime", MFnNumericData::kDouble, 0.0);
	nAttr.setStorable(false);

	aExclusiveCamera = mAttr.create("exclusiveCamera", "ecm");

	aSceneStateTracking = eAttr.create("sceneStateTracking", "sst", MFnNumericData::kShort);
	eAttr.addField("active", 0);
	eAttr.addField("passive", 1);
	eAttr.setDefault(0);
	eAttr.setKeyable(true);

	aReset = nAttr.create("reset", "reset", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aCull = nAttr.create("backfaceCulling", "backfaceCulling", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aSpecClrR = nAttr.create("specularColorR", "scr", MFnNumericData::kFloat, 0.5);
	nAttr.setMin(0.0);
	nAttr.setKeyable(true);
	aSpecClrG = nAttr.create("specularColorG", "scg", MFnNumericData::kFloat, 0.5);
	nAttr.setMin(0.0);
	nAttr.setKeyable(true);
	aSpecClrB = nAttr.create("specularColorB", "scb", MFnNumericData::kFloat, 0.5);
	nAttr.setMin(0.0);
	nAttr.setKeyable(true);

	aIsFrameCached = nAttr.create("isFrameCached", "isFrameCached", MFnNumericData::kBoolean, false);
	nAttr.setWritable(false);
	nAttr.setStorable(false);

	addAttribute(aSceneStateTracking);
	addAttribute(aReset);
	addAttribute(aCull);
	addAttribute(aSpecClrR);
	addAttribute(aSpecClrG);
	addAttribute(aSpecClrB);
	addAttribute(aInBaseAttr);
	addAttribute(aInData);
	addAttribute(aInTime);
	addAttribute(aExclusiveCamera);
	addAttribute(aIsFrameCached);

	attributeAffects(aInBaseAttr, aIsFrameCached);
	attributeAffects(aInAttr, aIsFrameCached);
	attributeAffects(aInTime, aIsFrameCached);
	attributeAffects(aSceneStateTracking, aIsFrameCached);
	attributeAffects(aReset, aIsFrameCached);
	attributeAffects(aCull, aIsFrameCached);
	attributeAffects(aSpecClrR, aIsFrameCached);
	attributeAffects(aSpecClrG, aIsFrameCached);
	attributeAffects(aSpecClrB, aIsFrameCached);

	return MS::kSuccess;
}

MStatus gpuCacher::compute(const MPlug& plug, MDataBlock& data)
{
	if(plug == aIsFrameCached)
	{
		if(!_glew)
		{
			data.outputValue(aIsFrameCached).set(false);
			return MS::kSuccess;
		}

		_oThisNode = thisMObject();

		if(MPlug(_oThisNode, aReset).asBool()) reset();

		_ifc = true; // flag - is frame cached
		_cull = data.inputValue(aCull).asBool();
		_t = data.inputValue(aInTime).asDouble(); // input time

		if(data.inputValue(aSceneStateTracking).asShort() <= 0) // if active scene state tracking
		{
			MArrayDataHandle adh = data.inputArrayValue(aInBaseAttr);
			cacheBaseAttr(adh); // update base input attributes cache
			cacheAttr(); // update per-group input attributes cache
			cacheGeo(); // update per-group mesh data cache
		}
		else // if passive scene state tracking
		{
			if(_mapGeo.find(_t) == _mapGeo.end()) // if frame is not cached
			{
				MArrayDataHandle adh = data.inputArrayValue(aInBaseAttr);
				cacheBaseAttr(adh); // update base input attributes cache
				cacheAttr(); // update per-group input attributes cache
				cacheGeo(); // update per-group mesh data cache
			}
			else // if frame is cached
			{
				_nCachedGeoCnt = _mapGeo[_t].length();
				_nCachedIdsCnt = _mapIds[_t].length();
			}
		}

		_scr = data.inputValue(aSpecClrR).asFloat();
		_scg = data.inputValue(aSpecClrG).asFloat();
		_scb = data.inputValue(aSpecClrB).asFloat();

		// update vbo
		//
		_geo = new GLfloat[_nCachedGeoCnt]; _mapGeo[_t].get(_geo);
		_ids = new GLuint[_nCachedIdsCnt]; _mapIds[_t].get(_ids);

		_nCachedGeoCnt /= 10;
		if(_nCachedGeoCnt != _nGeoCnt || _nCachedIdsCnt != _nIdsCnt)
		{
			_nGeoCnt = _nCachedGeoCnt;
			_nIdsCnt = _nCachedIdsCnt;
			vbo_init();
		}

		vbo_update();

		delete [] _geo;
		delete [] _ids;
		//
		// end update vbo

		// update output attribute
		data.outputValue(aIsFrameCached).set(_ifc);
	}

	return MS::kSuccess;
}

void gpuCacher::draw(M3dView &view, const MDagPath &path, M3dView::DisplayStyle style, M3dView::DisplayStatus status)
{
	// if exclusive camera provided and not the one of this view - don't draw anything
	MPlug plugExclusiveCamera(_oThisNode, aExclusiveCamera);
	if(plugExclusiveCamera.isConnected())
	{
		MDagPath dagPathThisCamera;
		view.getCamera(dagPathThisCamera);
		MPlugArray pa;
		plugExclusiveCamera.connectedTo(pa, true, false);
		MDagPath dagPathExclusiveCamera;
		MFnDagNode(pa[0].node()).getPath(dagPathExclusiveCamera);
		if(dagPathExclusiveCamera.fullPathName() != dagPathThisCamera.fullPathName()) return;
	}

	MPlug(thisMObject(), aIsFrameCached).asBool();

	if(!_glew || !_ifc || _nGeoCnt == 0 || _nIdsCnt == 0) return;

	view.beginGL();
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, _vboGeo);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, _vboIds);

//	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);

//	glTexCoordPointer(3, GL_FLOAT, sizeof(GLfloat)*10, (float*)(sizeof(GLfloat)*10));
	glNormalPointer(GL_FLOAT, sizeof(GLfloat)*10, (float*)(sizeof(GLfloat)*3));
	glVertexPointer(3, GL_FLOAT, sizeof(GLfloat)*10, NULL);

	if(_cull)
	{
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);
	}

	if(style == M3dView::kPoints)
	{
		glPointSize(4);
		glDrawElements(GL_POINTS, _nIdsCnt, GL_UNSIGNED_INT, NULL);
	}
	else if(style == M3dView::kWireFrame)
	{
		if(_cull) glPolygonMode(GL_FRONT, GL_LINE);
		else glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDrawElements(GL_TRIANGLES, _nIdsCnt, GL_UNSIGNED_INT, NULL);
	}
	else
	{
		if(view.wireframeOnShaded() || status == M3dView::kHilite ||
			status == M3dView::kLead || status == M3dView::kActive)
		{
			int nClrId = 4;
			switch(status)
			{
				case M3dView::kHilite:
					nClrId = 17;
					break;

				case M3dView::kLead:
					nClrId = 18;
					break;

				case M3dView::kActive:
					nClrId = 15;
					break;

				case M3dView::kActiveAffected:
					nClrId = 8;
					break;

				case M3dView::kTemplate:
					nClrId = 2;
					break;

				case M3dView::kActiveTemplate:
					nClrId = 19;
					break;

				default:
					nClrId = 4;
			}
			glLineWidth(2.0);
			view.setDrawColor(nClrId, M3dView::kDormantColors);
			if(_cull) glPolygonMode(GL_FRONT, GL_LINE);
			else glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glDrawElements(GL_TRIANGLES, _nIdsCnt, GL_UNSIGNED_INT, NULL);
		}

		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4, GL_FLOAT, sizeof(GLfloat)*10, (float*)(sizeof(GLfloat)*6));

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glShadeModel(GL_SMOOTH);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		GLfloat specReflection[] = {_scr, _scg, _scb, 1.0};
		glMaterialfv(GL_FRONT, GL_SPECULAR, specReflection);
		glMateriali(GL_FRONT, GL_SHININESS, 128);
		glEnable(GL_CULL_FACE);
		if(_cull) glPolygonMode(GL_FRONT, GL_FILL);
		else glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glDrawElements(GL_TRIANGLES, _nIdsCnt, GL_UNSIGNED_INT, NULL);

		glDisableClientState(GL_COLOR_ARRAY);
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
//	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	glPopAttrib();
	view.endGL();

	return;
}

void gpuCacher::vbo_create()
{
	glGenBuffersARB(1, &_vboGeo);
	glGenBuffersARB(1, &_vboIds);
}

void gpuCacher::vbo_delete()
{
	glDeleteBuffersARB(1, &_vboGeo);
	glDeleteBuffersARB(1, &_vboIds);
}

void gpuCacher::vbo_init()
{
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, _vboGeo);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, sizeof(GLfloat)*10*_nGeoCnt, NULL, GL_DYNAMIC_DRAW_ARB);
	cout<<"VBO init status: "<<glGetError()<<endl;
	if(glGetError() != GL_NO_ERROR)
	{
		cout<<"VBO initialization failed."<<endl;
		_nGeoCnt = _nIdsCnt = 0;
		return;
	}

	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, _vboIds);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, sizeof(GLuint)*_nIdsCnt, NULL, GL_STATIC_DRAW_ARB);
	if(glGetError() != GL_NO_ERROR)
	{
		cout<<"VBO initialization failed."<<endl;
		_nGeoCnt = _nIdsCnt = 0;
		return;
	}
}

void gpuCacher::vbo_update()
{
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, _vboGeo);
	glBufferSubDataARB(GL_ARRAY_BUFFER_ARB, 0, sizeof(GLfloat)*10*_nGeoCnt, _geo);

	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, _vboIds);
	glBufferSubDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0, sizeof(GLuint)*_nIdsCnt, _ids);
}

void gpuCacher::cacheBaseAttr(MArrayDataHandle& adh)
{
	float f = 0.0;
	MFloatArray fa(_mapBaseAttr[_t]);
	unsigned int cnt = adh.elementCount();

	if(fa.length() != cnt)
	{
		fa.setLength(cnt);
		_ifc = false;
	}

	for(unsigned int i = 0; i < cnt; ++i)
	{
		adh.jumpToArrayElement(i);
		f = adh.inputValue().asFloat();
		if(f != fa[i])
		{
			_ifc = false;
			fa[i] = f;
		}
	}

	_mapBaseAttr[_t] = fa;
}

void gpuCacher::cacheAttr()
{
	float f = 0.0;
	unsigned int i, j, cnt; i = j = cnt = 0;
	MPlug pInAttr;
	MPlug pInData(_oThisNode, aInData);
	for(i = 0; i < pInData.numElements(); ++i)
	{
		pInAttr = pInData.elementByPhysicalIndex(i).child(0);
		cnt = pInAttr.numElements();

		_mapIfc[i][_t] = true;
		MFloatArray fa(_mapAttr[i][_t]);
		if(fa.length() != cnt)
		{
			fa.setLength(cnt);
			_ifc = _mapIfc[i][_t] = false;
		}

		for(j = 0; j < cnt; ++j)
		{
			f = pInAttr.elementByPhysicalIndex(j).asFloat();
			if(f != fa[j])
			{
				_ifc = _mapIfc[i][_t] = false;
				fa[j] = f;
			}
		}

		_mapAttr[i][_t] = fa;
	}
}

void gpuCacher::cacheGeo()
{
	MFloatArray fArrGeo;
	MUintArray nArrIds;
	MUintArray nArrRng;

	// check if cached data exists for this frame
	bool igc = true;
	if(_mapRng.find(_t) == _mapRng.end())
	{
		_ifc = false;
		igc = false;
		_mapGeo[_t].clear();
		_mapIds[_t].clear();
		_mapRng[_t].clear();
	}

	// if any of the base input attributes changed or
	// there is no geometry cache for this frame
	if(!_ifc || !igc)
	{
		unsigned int i, j, k, idx, cnt, min, max, min2, max2, min3, nPntCnt, nIdsCnt, nClrCnt;
		i = j = k = idx = cnt = min = max = min2 = max2 = min3 = nPntCnt = nIdsCnt = nClrCnt = 0;
		double r, g, b, a; r = g = b = a = 0.0;
		MObject o;
		MPlug p2, p3, p4;
		MIntArray na, nArrIds2;
		MFloatPointArray pArrPos;
		MFloatVectorArray vArrNml;
		MPlug p(_oThisNode, aInData);
		fArrGeo = _mapGeo[_t];
		nArrIds = _mapIds[_t];
		nArrRng = _mapRng[_t];
		for(i = 0; i < p.numElements(); ++i)
		{
			// if any of the input attributes for the current array changed
			if(!igc) _mapIfc[i][_t] = false;
			if(!_ifc || !_mapIfc[i][_t])
			{
				// for each mesh in the current array
				p2 = p.elementByPhysicalIndex(i);

				o = p2.child(2).asMObject();
				MFnNumericData(o).getData4Double(r, g, b, a);

				p2 = p2.child(1);
				for(j = 0; j < p2.numElements(); ++j)
				{
					p3 = p2.elementByPhysicalIndex(j);
					if(!p3.isConnected()) continue;

					// read mesh data
					o = p3.asMObject();
					MFnMesh fnMesh(o);
					fnMesh.getTriangles(na, nArrIds2);
					fnMesh.getPoints(pArrPos, MSpace::kWorld);
					fnMesh.getVertexNormals(false, vArrNml, MSpace::kWorld);
					nPntCnt = fnMesh.numVertices();
					nIdsCnt = nArrIds2.length();

					// calculate min/max buffer offsets
					min = 0;
					max = 0;
					min2 = 0;
					max2 = 0;
					cnt = nArrRng.length();

					if(idx*4 < cnt)
					{
						min = nArrRng[idx*4];
						max = nArrRng[idx*4+1];
						min2 = nArrRng[idx*4+2];
						max2 = nArrRng[idx*4+3];
						if(nPntCnt != max-min || nIdsCnt != max2-min2)
						{
							cout<<"gpuCacher: Reset internal cache for topology change at inMeshList["<<i<<"].mesh["<<j<<"]."<<endl;
							reset();
							return;
						}
					}
					else
					{
						if(cnt == 0)
						{
							min = 0;
							max = nPntCnt;
							min2 = 0;
							max2 = nIdsCnt;
						}
						else
						{
							min = nArrRng[cnt-3];
							max = min+nPntCnt;
							min2 = nArrRng[cnt-1];
							max2 = min2+nIdsCnt;
						}

						// update min/max offsets
						nArrRng.append(min);
						nArrRng.append(max);
						nArrRng.append(min2);
						nArrRng.append(max2);

						// expand buffers
						fArrGeo.setLength(max*10);
						nArrIds.setLength(max2);
					}

					// update buffers
					int kk = 0;
					min3 = min*10;
					for(k = 0; k < nPntCnt; ++k)
					{
						kk = min3+k*10;
						fArrGeo[kk] = pArrPos[k].x;
						fArrGeo[kk+1] = pArrPos[k].y;
						fArrGeo[kk+2] = pArrPos[k].z;
						fArrGeo[kk+3] = vArrNml[k].x;
						fArrGeo[kk+4] = vArrNml[k].y;
						fArrGeo[kk+5] = vArrNml[k].z;
						fArrGeo[kk+6] = r;
						fArrGeo[kk+7] = g;
						fArrGeo[kk+8] = b;
						fArrGeo[kk+9] = a;
//						fArrGeo[min3+k*12+10] = 0.0;
//						fArrGeo[min3+k*12+11] = 0.0;
					}

					for(k = 0; k < nIdsCnt; ++k)
						nArrIds[min2+k] = min+nArrIds2[k];

					++idx;
				}
				_mapIfc[i][_t] = true;
			}
		}

		_mapGeo[_t] = fArrGeo;
		_mapIds[_t] = nArrIds;
		_mapRng[_t] = nArrRng;
	}

	if(_ifc) _nCachedGeoCnt = _mapGeo[_t].length();
	else _nCachedGeoCnt = fArrGeo.length();

	if(_ifc) _nCachedIdsCnt = _mapIds[_t].length();
	else _nCachedIdsCnt = nArrIds.length();

	_ifc = true;
}

void gpuCacher::reset()
{
	_ifc = false;
	_mapIfc.clear();
	_mapBaseAttr.clear();
	_mapAttr.clear();
	_mapGeo.clear();
	_mapIds.clear();
	_mapRng.clear();

	_nGeoCnt = _nIdsCnt = _nCachedGeoCnt = _nCachedIdsCnt = 0;

	MPlug(_oThisNode, aReset).setBool(false);
}



MString gpuCacher::drawDbClassification("drawdb/geometry/gpuCacher");
MString gpuCacher::drawRegistrantId("GpuCacherNodePlugin");

GpuCacherDrawOverride::GpuCacherDrawOverride(const MObject& obj)
: MHWRender::MPxDrawOverride(obj, GpuCacherDrawOverride::draw)
{
}

GpuCacherDrawOverride::~GpuCacherDrawOverride()
{
}

MHWRender::DrawAPI GpuCacherDrawOverride::supportedDrawAPIs() const
{
	return (MHWRender::kOpenGL | MHWRender::kOpenGLCoreProfile);
}

MUserData* GpuCacherDrawOverride::prepareForDraw(
	const MDagPath& objPath,
	const MDagPath& cameraPath,
	const MHWRender::MFrameContext& frameContext,
	MUserData* oldData)
{
	GpuCacherData* data = dynamic_cast<GpuCacherData*>(oldData);
	if(!data) data = new GpuCacherData();

	data->dpObj = objPath;
	data->dpCam = cameraPath;

	return data;
}

void GpuCacherDrawOverride::addUIDrawables(
		const MDagPath& objPath,
		MHWRender::MUIDrawManager& drawManager,
		const MHWRender::MFrameContext& frameContext,
		const MUserData* data)
{
	GpuCacherData* d = (GpuCacherData*)data; if(!d) return;

	drawManager.beginDrawable();
	drawManager.endDrawable();
}

void GpuCacherDrawOverride::draw(const MHWRender::MDrawContext& context, const MUserData* data)
{
	const GpuCacherData* d = dynamic_cast<const GpuCacherData*>(data); if(!d) return;
	MHWRender::MRenderer* r = MHWRender::MRenderer::theRenderer(); if(!r) return;
	const unsigned int ds = context.getDisplayStyle(); if(ds == 0) return;

	MObject o = d->dpObj.node();
	gpuCacher* ptrGpuCacher = dynamic_cast<gpuCacher*>(MFnDependencyNode(o).userNode());

	MPlug(o, ptrGpuCacher->aIsFrameCached).asBool(); // force evaluate

	// set world view matrix
	MMatrix mWorldView = context.getMatrix(MHWRender::MFrameContext::kWorldViewMtx);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadMatrixd(mWorldView.matrix[0]);
	// set projection matrix
	MMatrix mProjection = context.getMatrix(MHWRender::MFrameContext::kProjectionMtx);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadMatrixd(mProjection.matrix[0]);
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	// if exclusive camera provided and not the one of this view - don't draw anything
	MPlug p(o, ptrGpuCacher->aExclusiveCamera);
	if(p.isConnected())
	{
		MPlugArray pa;
		p.connectedTo(pa, true, false);
		MDagPath dp;
		MFnDagNode(pa[0].node()).getPath(dp);
		if(dp.fullPathName() != d->dpCam.fullPathName()) return;
	}

	if(!ptrGpuCacher->_glew || !ptrGpuCacher->_ifc || ptrGpuCacher->_nGeoCnt == 0 || ptrGpuCacher->_nIdsCnt == 0) return;

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, ptrGpuCacher->_vboGeo);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, ptrGpuCacher->_vboIds);

//	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);

//	glTexCoordPointer(3, GL_FLOAT, sizeof(GLfloat)*10, (float*)(sizeof(GLfloat)*10));
	glNormalPointer(GL_FLOAT, sizeof(GLfloat)*10, (float*)(sizeof(GLfloat)*3));
	glVertexPointer(3, GL_FLOAT, sizeof(GLfloat)*10, NULL);

	if(ptrGpuCacher->_cull)
	{
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);
	}

	MColor c = MHWRender::MGeometryUtilities::wireframeColor(d->dpObj);
	glColor3f(c.r, c.g, c.b);

	if(!(ds & MHWRender::MFrameContext::kGouraudShaded))
	{
		// draw geometry as wireframe
		if(ptrGpuCacher->_cull) glPolygonMode(GL_FRONT, GL_LINE);
		else glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDrawElements(GL_TRIANGLES, ptrGpuCacher->_nIdsCnt, GL_UNSIGNED_INT, NULL);
	}
	else
	{
		if(ds & MHWRender::MFrameContext::kWireFrame || MHWRender::MGeometryUtilities::displayStatus(d->dpObj) != MHWRender::kDormant)
		{
			// draw geometry as wireframe
			glLineWidth(2.0);
			if(ptrGpuCacher->_cull) glPolygonMode(GL_FRONT, GL_LINE);
			else glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glDrawElements(GL_TRIANGLES, ptrGpuCacher->_nIdsCnt, GL_UNSIGNED_INT, NULL);
		}



		// handle Maya lights
		unsigned int nLights = std::min(context.numberOfActiveLights(), 8u);
		if(nLights > 0)
			for (unsigned int i = 0; i < nLights; ++i)
			{
				MFloatPointArray positions;
				MFloatVector direction;
				float intensity;
				MColor color;
				bool hasDirection;
				bool hasPosition;
				context.getLightInformation(i, positions, direction, intensity, color, hasDirection, hasPosition);

				if(hasDirection)
				{
					if(hasPosition)
					{
						// spot light
						MFloatPoint position;
						position[0] = positions[0][0];
						position[1] = positions[0][1];
						position[2] = positions[0][2];
						const GLfloat ambient[4] = {0.0f, 0.0f, 0.0f, 1.0f};
						const GLfloat diffuse[4] = {intensity * color[0], intensity * color[1], intensity * color[2], 1.0f};
						const GLfloat pos[4] = {position[0], position[1], position[2], 1.0f};
						const GLfloat dir[3] = {direction[0], direction[1], direction[2]};

						glLightfv(GL_LIGHT0+i, GL_AMBIENT, ambient);
						glLightfv(GL_LIGHT0+i, GL_DIFFUSE, diffuse);
						glLightfv(GL_LIGHT0+i, GL_POSITION, pos);
						glLightfv(GL_LIGHT0+i, GL_SPOT_DIRECTION, dir);
						glLightf(GL_LIGHT0+i,  GL_SPOT_EXPONENT, 0.0);
						glLightf(GL_LIGHT0+i,  GL_SPOT_CUTOFF,  20.0);
					}
					else {
						// directional light
						const GLfloat ambient[4] = {0.0f, 0.0f, 0.0f, 1.0f};
						const GLfloat diffuse[4] = {intensity * color[0], intensity * color[1], intensity * color[2], 1.0f};
						const GLfloat pos[4] = {-direction[0], -direction[1], -direction[2], 0.0f};

						glLightfv(GL_LIGHT0+i, GL_AMBIENT,  ambient);
						glLightfv(GL_LIGHT0+i, GL_DIFFUSE,  diffuse);
						glLightfv(GL_LIGHT0+i, GL_POSITION, pos);
						glLightf(GL_LIGHT0+i, GL_SPOT_CUTOFF, 180.0);
					}
				}
				else if(hasPosition)
				{
					// point light
					MFloatPoint position;
					position[0] = positions[0][0];
					position[1] = positions[0][1];
					position[2] = positions[0][2];
					const GLfloat ambient[4] = {0.0f, 0.0f, 0.0f, 1.0f};
					const GLfloat diffuse[4] = {intensity * color[0], intensity * color[1], intensity * color[2], 1.0f};
					const GLfloat pos[4] = {position[0], position[1], position[2], 1.0f};

					glLightfv(GL_LIGHT0+i, GL_AMBIENT,  ambient);
					glLightfv(GL_LIGHT0+i, GL_DIFFUSE,  diffuse);
					glLightfv(GL_LIGHT0+i, GL_POSITION, pos);
					glLightf(GL_LIGHT0+i, GL_SPOT_CUTOFF, 180.0);
				}
				else
				{
					// ambient light
					const GLfloat ambient[4] = {intensity * color[0], intensity * color[1], intensity * color[2], 1.0f};
					const GLfloat diffuse[4] = {0.0f, 0.0f, 0.0f, 1.0f};
					const GLfloat pos[4] = {0.0f, 0.0f, 0.0f, 1.0f};

					glLightfv(GL_LIGHT0+i, GL_AMBIENT,  ambient);
					glLightfv(GL_LIGHT0+i, GL_DIFFUSE,  diffuse);
					glLightfv(GL_LIGHT0+i, GL_POSITION, pos);
					glLightf(GL_LIGHT0+i, GL_SPOT_CUTOFF, 180.0);
				}

				glEnable(GL_LIGHT0+i);
			}



		// draw geometry
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4, GL_FLOAT, sizeof(GLfloat)*10, (float*)(sizeof(GLfloat)*6));

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glShadeModel(GL_SMOOTH);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		GLfloat specReflection[] = {ptrGpuCacher->_scr, ptrGpuCacher->_scg, ptrGpuCacher->_scb, 1.0};
		glMaterialfv(GL_FRONT, GL_SPECULAR, specReflection);
		glMateriali(GL_FRONT, GL_SHININESS, 128);
		glEnable(GL_CULL_FACE);
		if(ptrGpuCacher->_cull) glPolygonMode(GL_FRONT, GL_FILL);
		else glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glDrawElements(GL_TRIANGLES, ptrGpuCacher->_nIdsCnt, GL_UNSIGNED_INT, NULL);

		glDisableClientState(GL_COLOR_ARRAY);



		// handle Maya lights
		if(nLights > 0)
			for(unsigned int i = 0; i < nLights; ++i)
			{
				glDisable(GL_LIGHT0+i);
				const GLfloat ambient[4] = {0.0f, 0.0f, 0.0f, 1.0f};
				glLightfv(GL_LIGHT0+i, GL_AMBIENT, ambient);
				if(i == 0)
				{
					const GLfloat diffuse[4] = {1.0f, 1.0f, 1.0f, 1.0f};
					glLightfv(GL_LIGHT0+i, GL_DIFFUSE, diffuse);
					const GLfloat spec[4] = {1.0f, 1.0f, 1.0f, 1.0f};
					glLightfv(GL_LIGHT0+i, GL_SPECULAR, spec);
				}
				else
				{
					const GLfloat diffuse[4] = {0.0f, 0.0f, 0.0f, 1.0f};
					glLightfv(GL_LIGHT0+i, GL_DIFFUSE, diffuse);
					const GLfloat spec[4] = {0.0f, 0.0f, 0.0f, 1.0f};
					glLightfv(GL_LIGHT0+i, GL_SPECULAR, spec);
				}
				const GLfloat pos[4] = {0.0f, 0.0f, 1.0f, 0.0f};
				glLightfv(GL_LIGHT0+i, GL_POSITION, pos);
				const GLfloat dir[3] = {0.0f, 0.0f, -1.0f};
				glLightfv(GL_LIGHT0+i, GL_SPOT_DIRECTION, dir);
				glLightf(GL_LIGHT0+i, GL_SPOT_EXPONENT, 0.0);
				glLightf(GL_LIGHT0+i, GL_SPOT_CUTOFF, 180.0);
			}
	}



	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
//	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	glPopAttrib();
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}
