#if __linux || _win32
#include <omp.h>
#endif

#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>
#include <maya/MGlobal.h>
#include <maya/MPlug.h>
#include <maya/MPlugArray.h>
#include <maya/MObject.h>
#include <maya/MTypeId.h>
#include <maya/MFnNumericAttribute.h>
#include <maya/MFnGenericAttribute.h>
#include <maya/MFnTypedAttribute.h>
#include <maya/MFnEnumAttribute.h>
#include <maya/MFnMesh.h>
#include <maya/MFnMeshData.h>
#include <maya/MVectorArray.h>
#include <maya/MFnVectorArrayData.h>
#include <maya/MSelectionList.h>
#include <maya/MMeshIntersector.h>
#include <maya/MString.h>
#include <maya/MStringArray.h>

#include "remesh.h"
#include "../pluginMain.h"



#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <vector>
#include <fstream>
typedef CGAL::Exact_predicates_inexact_constructions_kernel K;

#if __linux__ || _WIN32
// cgal
#include <CGAL/Surface_mesh.h>
#include <CGAL/boost/graph/graph_traits_Surface_mesh.h>
#include <CGAL/Polygon_mesh_processing/remesh.h>
#include <CGAL/Polygon_mesh_processing/border.h>
#include <boost/function_output_iterator.hpp>

typedef CGAL::Surface_mesh<K::Point_3> Mesh;
typedef Mesh::Vertex_index vertex_descriptor;
typedef boost::graph_traits<Mesh>::halfedge_descriptor halfedge_descriptor;
typedef boost::graph_traits<Mesh>::edge_descriptor     edge_descriptor;
namespace PMP = CGAL::Polygon_mesh_processing;

struct halfedge2edge
{
  halfedge2edge(const Mesh& m, std::vector<edge_descriptor>& edges) : m_mesh(m), m_edges(edges) {}
  void operator()(const halfedge_descriptor& h) const
  {
    m_edges.push_back(edge(h, m_mesh));
  }
  const Mesh& m_mesh;
  std::vector<edge_descriptor>& m_edges;
};
#endif



#include <CGAL/Scale_space_surface_reconstruction_3.h>
typedef CGAL::Scale_space_surface_reconstruction_3<K>           Reconstruction;
typedef Reconstruction::Point                                   Point;
typedef std::vector<Point>                                      Point_collection;
typedef Reconstruction::Triple_const_iterator                   Triple_iterator;



MTypeId remesh::id(ID_REMESH);
MObject remesh::aInMesh;
MObject remesh::aInPosPP;
MObject remesh::aInNmlPP;
MObject remesh::aOutMesh;
MObject remesh::aActive;
MObject remesh::aMode;
MObject remesh::aUseMsh;
MObject remesh::aUseCld;
MObject remesh::aPreSubd;
MObject remesh::aDepth;
MObject remesh::aScale;
MObject remesh::aSamplesPerNode;
MObject remesh::aPointWeight;
MObject remesh::aThreads;
MObject remesh::aUseNormalWeights;
MObject remesh::aNonManifold;
MObject remesh::aTriangulate;
MObject remesh::aAdaptiveExponent;
MObject remesh::aPreserveBoundary;
MObject remesh::aFullDepth;
MObject remesh::aIters;
MObject remesh::aConform;
MObject remesh::aEdgeLength;
MObject remesh::aUniformity;
MObject remesh::aEdgeLength2;
MObject remesh::aPreserveDetail;
MObject remesh::aNeighbors;
MObject remesh::aSamples;
MObject remesh::aIters2;

remesh::remesh(){}
remesh::~remesh(){}

void* remesh::creator()
{
	return new remesh();
}

MStatus remesh::initialize()
{
	MFnNumericAttribute nAttr;
	MFnGenericAttribute gAttr;
	MFnTypedAttribute tAttr;
	MFnEnumAttribute eAttr;

	aInMesh = gAttr.create("inMesh", "im");
	gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aInPosPP = gAttr.create("inPositionPP", "ippp");
	gAttr.addDataAccept(MFnData::kVectorArray);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aInNmlPP = gAttr.create("inNormalPP", "inpp");
	gAttr.addDataAccept(MFnData::kVectorArray);
	gAttr.setStorable(false);
	gAttr.setDisconnectBehavior(MFnAttribute::kReset);

	aOutMesh = gAttr.create("outMesh", "om");
	gAttr.addDataAccept(MFnData::kMesh);
	gAttr.setWritable(false);
	gAttr.setStorable(false);

	aActive = nAttr.create("active", "act", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aMode = eAttr.create("mode", "mode", MFnNumericData::kShort);
	eAttr.addField("poisson", 0);
	eAttr.addField("isotropic", 1);
	eAttr.addField("topology", 2);
	eAttr.addField("unoriented points", 3);
	eAttr.setDefault(0);
	eAttr.setKeyable(true);

	// poisson
	//
	aUseMsh = nAttr.create("useBoundaryMesh", "ubm", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aUseCld = nAttr.create("usePointCloud", "upc", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aThreads = nAttr.create("threads", "threads", MFnNumericData::kInt, 8);
	nAttr.setKeyable(true);
	nAttr.setMin(1);
	nAttr.setSoftMax(12);

	aPreSubd = nAttr.create("preSubdivide", "preSubdivide", MFnNumericData::kInt, 0);
	nAttr.setKeyable(true);
	nAttr.setMin(0);
	nAttr.setSoftMax(2);

	aScale = nAttr.create("scale", "scale", MFnNumericData::kFloat, 1.1);
	nAttr.setKeyable(true);

	aDepth = nAttr.create("depth", "depth", MFnNumericData::kInt, 4);
	nAttr.setKeyable(true);
	nAttr.setMin(0);
	nAttr.setSoftMax(4);

	aSamplesPerNode = nAttr.create("samplesPerNode", "samplesPerNode", MFnNumericData::kFloat, 1.5);
	nAttr.setKeyable(true);

	aPointWeight = nAttr.create("pointWeight", "pointWeight", MFnNumericData::kFloat, 4.0);
	nAttr.setKeyable(true);

	aUseNormalWeights = nAttr.create("useNormalWeights", "useNormalWeights", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aNonManifold = nAttr.create("nonManifold", "nonManifold", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aTriangulate = nAttr.create("triangulate", "triangulate", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	aAdaptiveExponent = nAttr.create("adaptiveExponent", "adaptiveExponent", MFnNumericData::kInt, 1);
	nAttr.setKeyable(true);
	nAttr.setMin(0);
	nAttr.setSoftMax(3);

	aPreserveBoundary = nAttr.create("preserveBoundary", "preserveBoundary", MFnNumericData::kBoolean, true);
	nAttr.setKeyable(true);

	aFullDepth = nAttr.create("fullDepth", "fullDepth", MFnNumericData::kInt, 5);
	nAttr.setKeyable(true);
	nAttr.setMin(0);

	aIters = nAttr.create("iterations", "iterations", MFnNumericData::kInt, 4);
	nAttr.setKeyable(true);
	nAttr.setMin(1);

	aConform = nAttr.create("conform", "conform", MFnNumericData::kBoolean, false);
	nAttr.setKeyable(true);

	// isotropic
	//
	aEdgeLength = nAttr.create("edgeLength", "el", MFnNumericData::kDouble, 1.0);
	nAttr.setSoftMin(0.1);
	nAttr.setSoftMax(2.0);
	nAttr.setKeyable(true);

	aUniformity = nAttr.create("uniformity", "uni", MFnNumericData::kInt, 3);
	nAttr.setMin(0);
	nAttr.setKeyable(true);

	// topology
	//
	aEdgeLength2 = nAttr.create("edgeLength2", "el2", MFnNumericData::kDouble, 1.0);
	nAttr.setSoftMin(0.1);
	nAttr.setSoftMax(2.0);
	nAttr.setKeyable(true);

	aPreserveDetail = nAttr.create("preserveDetail", "pdet", MFnNumericData::kDouble, 100.0);
	nAttr.setMin(0.0);
	nAttr.setMax(100.0);
	nAttr.setKeyable(true);

	// topology
	//
	aNeighbors = nAttr.create("neighbors", "nbrs", MFnNumericData::kInt, 20);
	nAttr.setMin(1);
	nAttr.setSoftMax(100);
	nAttr.setKeyable(true);

	aSamples = nAttr.create("samples", "smp", MFnNumericData::kInt, 100);
	nAttr.setMin(1);
	nAttr.setSoftMax(200);
	nAttr.setKeyable(true);

	aIters2 = nAttr.create("iterations2", "iterations2", MFnNumericData::kInt, 1);
	nAttr.setKeyable(true);
	nAttr.setMin(1);
	nAttr.setSoftMax(5);

	addAttribute(aInMesh);
	addAttribute(aInPosPP);
	addAttribute(aInNmlPP);
	addAttribute(aOutMesh);
	addAttribute(aActive);
	addAttribute(aMode);
	addAttribute(aUseMsh);
	addAttribute(aUseCld);
	addAttribute(aMode);
	addAttribute(aThreads);
	addAttribute(aPreSubd);
	addAttribute(aDepth);
	addAttribute(aScale);
	addAttribute(aSamplesPerNode);
	addAttribute(aPointWeight);
	addAttribute(aUseNormalWeights);
	addAttribute(aNonManifold);
	addAttribute(aTriangulate);
	addAttribute(aAdaptiveExponent);
	addAttribute(aPreserveBoundary);
	addAttribute(aFullDepth);
	addAttribute(aIters);
	addAttribute(aConform);
	addAttribute(aEdgeLength);
	addAttribute(aUniformity);
	addAttribute(aEdgeLength2);
	addAttribute(aPreserveDetail);
	addAttribute(aNeighbors);
	addAttribute(aSamples);
	addAttribute(aIters2);

	attributeAffects(aInMesh, aOutMesh);
	attributeAffects(aInPosPP, aOutMesh);
	attributeAffects(aInNmlPP, aOutMesh);
	attributeAffects(aActive, aOutMesh);
	attributeAffects(aMode, aOutMesh);
	attributeAffects(aUseMsh, aOutMesh);
	attributeAffects(aUseCld, aOutMesh);
	attributeAffects(aThreads, aOutMesh);
	attributeAffects(aPreSubd, aOutMesh);
	attributeAffects(aDepth, aOutMesh);
	attributeAffects(aScale, aOutMesh);
	attributeAffects(aSamplesPerNode, aOutMesh);
	attributeAffects(aPointWeight, aOutMesh);
	attributeAffects(aUseNormalWeights, aOutMesh);
	attributeAffects(aNonManifold, aOutMesh);
	attributeAffects(aTriangulate, aOutMesh);
	attributeAffects(aAdaptiveExponent, aOutMesh);
	attributeAffects(aPreserveBoundary, aOutMesh);
	attributeAffects(aFullDepth, aOutMesh);
	attributeAffects(aIters, aOutMesh);
	attributeAffects(aConform, aOutMesh);
	attributeAffects(aEdgeLength, aOutMesh);
	attributeAffects(aUniformity, aOutMesh);
	attributeAffects(aEdgeLength2, aOutMesh);
	attributeAffects(aPreserveDetail, aOutMesh);
	attributeAffects(aNeighbors, aOutMesh);
	attributeAffects(aSamples, aOutMesh);
	attributeAffects(aIters2, aOutMesh);

	return MS::kSuccess;
}

MStatus remesh::compute(const MPlug& plug, MDataBlock& data)
{
	if(data.outputValue(state).asShort() == 2) return MS::kUnknownParameter;

	if(plug != aOutMesh) return MS::kUnknownParameter;

	MFnMeshData md;
	MObject o = md.create();
	MFnMesh m;
	m.copy(data.inputValue(aInMesh).asMesh(), o);
	m.setObject(o);

	if(data.inputValue(aActive).asBool() == false)
	{
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
	}

	int nPntCnt = m.numVertices();

	MStringArray sa;
	MPlug(thisMObject(), aInMesh).name().split('.', sa);

	short mode = data.inputValue(aMode).asShort();
	if(mode < 0) mode = 0;
	else if(mode > 3) mode = 3;

	if(mode == 0)
	{

	}
	else if(mode == 1)
	{
#ifdef __APPLE__
		MGlobal::displayWarning("This mode is not available on OSX.");
		MFnMesh m2;
		MFnMeshData md2;
		o = md2.create();
		m2.create(0, 0, MFloatPointArray(), MIntArray(), MIntArray(), o);
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
#else
		if(nPntCnt == 0)
		{
			MFnMesh m2;
			MFnMeshData md2;
			o = md2.create();
			m2.create(0, 0, MFloatPointArray(), MIntArray(), MIntArray(), o);
			data.outputValue(aOutMesh).set(o);
			return MS::kSuccess;
		}

		MFloatPointArray pnts;
		m.getPoints(pnts, MSpace::kWorld);

		Mesh mesh;

		std::vector<vertex_descriptor> vd(nPntCnt);
		for(unsigned int i = 0; i < nPntCnt; ++i)
			vd[i] = mesh.add_vertex(K::Point_3(pnts[i].x, pnts[i].y, pnts[i].z));

		MIntArray cnts, ids;
		m.getTriangles(cnts, ids);
		for(unsigned int i = 0; i < ids.length()/3; ++i)
			mesh.add_face(vd[ids[i*3]], vd[ids[i*3+1]], vd[ids[i*3+2]]);

		double el = data.inputValue(aEdgeLength).asDouble();
		if(el < 0.001) el = 0.001;
		unsigned int it = data.inputValue(aUniformity).asInt();
		if(it < 0) it = 0;

		std::vector<edge_descriptor> border;
		PMP::border_halfedges(faces(mesh), mesh, boost::make_function_output_iterator(halfedge2edge(mesh, border)));
		PMP::split_long_edges(border, el, mesh);
		PMP::isotropic_remeshing(faces(mesh), el, mesh, PMP::parameters::number_of_iterations(it).protect_constraints(true));
		mesh.collect_garbage();

		unsigned int i = 0;
		pnts.setLength(mesh.number_of_vertices());
		Mesh::Property_map<vertex_descriptor, K::Point_3> p = mesh.points();
		for(vertex_descriptor vd : mesh.vertices())
		{
			pnts[i].x = p[vd][0];
			pnts[i].y = p[vd][1];
			pnts[i].z = p[vd][2];
			pnts[i].w = 1;
			++i;
		}
		nPntCnt = i;

		unsigned int cnt = mesh.number_of_faces();
		cnts = MIntArray(cnt, 3);
		MIntArray conn(cnt*3);
		i = 0;
		for(Mesh::Face_iterator fit=mesh.faces_begin(); fit!=mesh.faces_end(); ++fit)
		{
			int nV = mesh.degree(*fit);
			Mesh::Vertex_around_face_circulator fvit(mesh.halfedge(*fit), mesh), fvend=fvit;
			do
			{
				Mesh::size_type idx = *fvit;
				conn[i] = idx;
				++i;
			}
			while (++fvit != fvend);
		}

		MFnMesh m2;
		MFnMeshData md2;
		o = md2.create();
		m2.create(nPntCnt, cnt, pnts, cnts, conn, o);
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
#endif
	}
	else if(mode == 2)
	{

	}
	else
	{
		MFloatPointArray pnts;
		if(nPntCnt) m.getPoints(pnts, MSpace::kWorld);
		MVectorArray va = MFnVectorArrayData(data.inputValue(aInPosPP).data()).array();
		unsigned int c = va.length();
		pnts.setLength(nPntCnt+c);
		for(unsigned int i = 0; i < c; ++i) pnts[nPntCnt+i] = va[i];
		nPntCnt += c;

		if(nPntCnt == 0)
		{
			MFnMesh m2;
			MFnMeshData md2;
			o = md2.create();
			m2.create(0, 0, MFloatPointArray(), MIntArray(), MIntArray(), o);
			data.outputValue(aOutMesh).set(o);
			return MS::kSuccess;
		}

		Point_collection points(nPntCnt);
		for(unsigned int i = 0; i < nPntCnt; ++i) points[i] = K::Point_3(pnts[i].x, pnts[i].y, pnts[i].z);

		int neighbors = data.inputValue(aNeighbors).asInt();
		if(neighbors < 1) neighbors = 1;
		int samples = data.inputValue(aSamples).asInt();
		if(samples < 1) samples = 1;
		int iterations = data.inputValue(aIters2).asInt()-1;
		if(iterations < 0) iterations = 0;

		Reconstruction reconstruct(neighbors, samples);
		reconstruct.reconstruct_surface(points.begin(), points.end(), iterations);

		c = reconstruct.number_of_triangles();
		MIntArray cnts(c);
		MIntArray conn(c*3);
		unsigned int i = 0;
		for(std::size_t shell = 0; shell < reconstruct.number_of_shells(); ++shell)
			for(Triple_iterator it = reconstruct.shell_begin(shell); it != reconstruct.shell_end(shell); ++it)
			{
				cnts[i] = 3;
				conn[i*3] = (*it)[0];
				conn[i*3+1] = (*it)[1];
				conn[i*3+2] = (*it)[2];
				++i;
			}

		MFnMesh m2;
		MFnMeshData md2;
		o = md2.create();
		m2.create(nPntCnt, c, pnts, cnts, conn, o);
		data.outputValue(aOutMesh).set(o);
		return MS::kSuccess;
	}
}
